import {
    AppModel,
    DerivedData,
    DerivedFunctions,
} from "../generated_module/src/types/AutoModuleTypes";
import * as Processor from "@microchip/scf-common/lib/Processor";
import { ClockCalculations } from "./ClockCalculations";
import * as i2c_host_basic from "@microchip/i2c-host-basic";
import * as i2c_master from "@microchip/i2c-host";
import * as I2CSlaveAVR8 from "./interfaces/I2CSlaveAVR8";
import { Arguments, Interface } from "@microchip/scf-interface";
import {
    getGeneratorModel,
    i2cMasterInterface,
    getI2CMasterTypes,
    i2cSlaveInterface,
} from "./GeneratorModel";
import { I2CModeFunctions, getI2CModeFunctions } from "./I2CMode";
import * as init_system from "@microchip/initializer-system";
import { intf_args } from "@microchip/pin-standard/lib/pin-standard";
import { MyComponentNames, MyExportNames } from "../generated_module/src/types";

export const getDerivedData = (dataModel: AppModel): DerivedData => {
    if (dataModel) {
        return new MyDerivedData(dataModel);
    }
    return new EmptyDerivedData();
};

//This data will be used at the creator stage and relies on only the state
class EmptyDerivedData implements DerivedData {
    getMyFunctions = (): DerivedFunctions => {
        return {};
    };
    getModel = (): AppModel | undefined => {
        return undefined;
    };
}

//This data will be used at the processor stage
class MyDerivedData implements DerivedData {
    private dataModel: AppModel;
    private clockCalc: ClockCalculations;

    constructor(dataModel: AppModel) {
        this.dataModel = dataModel;
    }

    private getCalc = (): ClockCalculations => {
        if (!this.clockCalc) {
            this.clockCalc = new ClockCalculations({
                clkPerInHz: Number(
                    this.dataModel?.getImportValue("osc_clocks")?.fosc ?? 4000000,
                ),
                fSclInHz: this.dataModel.getComponentValue("desiredClockSpeed") ?? 0,
                tRiseInUs: 0,
            });
        }
        return this.clockCalc;
    };

    getMyFunctions = (): DerivedFunctions => {
        return {
            overrideDefaultValues: this.overrideDefaultValues,
            moduleName: this.getModuleName,
            componentName: this.componentName,
            templateName: this.getTemplateName,
            interruptDriven: this.interruptDriven,
            interruptDrivenVisible: (): boolean =>
                this.dataModel.getComponentValue("mode") === "Host",
            isMasterModeInterrupt: (): boolean => this.isMasterModeInterrupt(),
            isMasterModePolling: (): boolean => this.isMasterModePolling(),
            isSlaveModeInterrupt: (): boolean => this.isSlaveModeInterrupt(),
            isMasterMode: (): boolean => this.isMasterMode(),
            shouldGenerateMasterExample: (): boolean =>
                this.shouldGenerateMasterExample(),
            "i2c-host-basic_payload": this.i2c_host_basic_payload,
            "i2c-host_payload": this.i2c_master_payload,
            "i2c-slave-avr8_payload": this.i2c_slave_avr8_payload,
            enableMctrla: this.enableMctrla,
            enableSctrla: this.enableSctrla,
            mbaud: this.mbaud,
            saddr: (): number | undefined =>
                this.dataModel.getComponentValue("slaveAddress"),
            actualClockSpeed: this.actualClockSpeed,
            pinData: this.getMyPinData,
            templateData: (): any => getGeneratorModel(this.dataModel),
            desiredClockSpeedValidator: this.desiredClockSpeedValidator,
            slaveAddressValidator: this.slaveAddressValidator,
            alerts: this.getMyAlerts,
            initializer_system_results: this.getInitializerSystemResults,
            interruptComponentName: this.interruptComponentName,
            pin_standard_args: this.getPinstandardArgs,
        };
    };

    private getPinstandardArgs = (): intf_args => {
        const args: intf_args = {
            queryAdapters: ["portmux"],
        };
        return args;
    };

    private getTemplateName = (src: string): string => {
        if (src === "output/i2c_host_poll.c.ftl" || src === "output/i2c_host_int.c.ftl") {
            return "i2c_host/src/" + this.getModuleName().toLowerCase() + ".c";
        } else if (
            src === "output/i2c_host_poll.h.ftl" ||
            src === "output/i2c_host_int.h.ftl"
        ) {
            return "i2c_host//" + this.getModuleName().toLowerCase() + ".h";
        } else if (src === "output/i2c_host_interface.h.ftl") {
            return (
                "i2c_host//" + this.getModuleName().toLowerCase() + "_host_interface.h"
            );
        } else if (src === "output/i2c_host_types.h.ftl") {
            return "i2c_host//" + this.getModuleName().toLowerCase() + "_host_types.h";
        } else if (src === "output/i2c_client_interrupt.c.ftl") {
            return "i2c_client/src/" + this.getModuleName().toLowerCase() + ".c";
        } else if (src === "output/i2c_client_interrupt.h.ftl") {
            return "i2c_client/" + this.getModuleName().toLowerCase() + ".h";
        } else {
            return "";
        }
    };

    private overrideDefaultValues = (componentName: string): any => {
        switch (componentName) {
            case "componentName":
                return (
                    this.dataModel.getHardware()?.getName().replace("TWI", "I2C") ?? ""
                );
        }
    };

    public getModel = (): AppModel => {
        return this.dataModel;
    };

    private getModeFunctions = (): I2CModeFunctions | undefined => {
        if (this.dataModel) {
            return getI2CModeFunctions(this.dataModel);
        }
        return undefined;
    };

    public getModuleName = (): string => {
        return this.dataModel.getHardware()?.getName() ?? "";
    };

    private componentName = (): string | undefined => {
        return (
            this.getFirstExportArg<i2c_master.Arguments>(i2c_master.Interface)
                ?.customName ??
            this.getFirstExportArg<I2CSlaveAVR8.Arguments>(I2CSlaveAVR8.Interface)
                ?.customName
        );
    };

    private interruptDriven = (): boolean | undefined => {
        if (this.dataModel.getComponentValue("mode") === "Client") return true;

        return (
            this.getFirstExportArg<i2c_master.Arguments>(i2c_master.Interface)
                ?.interruptDriven ??
            this.getFirstExportArg<I2CSlaveAVR8.Arguments>(I2CSlaveAVR8.Interface)
                ?.interruptDriven
        );
    };

    private getFirstExportArg = <T extends Arguments>(id: Interface): T | undefined => {
        const i2cExport = this.dataModel
            .getExportInterfaces()
            .getInterface("i2c", id.getInterfaceId());
        return i2cExport?.args?.[Object.keys(i2cExport.args)[0]];
    };

    private i2c_master_payload = (): i2c_master.ProcessedPayload => {
        const interfaceApi = i2cMasterInterface(this.dataModel);
        return {
            moduleName: this.dataModel.getName(),
            calculatedSpeed_Hz: this.dataModel.getComponentValue("actualClockSpeed") ?? 0,
            interfaceApi: interfaceApi,
            hasInterrupts: true,
            interruptDriven: this.dataModel.isComponentValue("interruptDriven"),
            typeHeader: interfaceApi?.headerFiles?.[0] ?? { name: "", path: "" },
            types: getI2CMasterTypes(this.dataModel),
        };
    };

    private i2c_slave_avr8_payload = (): I2CSlaveAVR8.ProcessedPayload => {
        return {
            interfaceApi: i2cSlaveInterface(this.dataModel),
            interruptDriven: this.dataModel.isComponentValue("interruptDriven"),
        };
    };

    private getInitializerSystemResults = (): init_system.Results => {
        const api = getGeneratorModel(this.dataModel);
        const header = api?.header;
        return {
            initializer: {
                name: api?.api?.Initialize?.name,
                header: {
                    name: header?.name,
                    path: header?.path?.replace("../", ""),
                },
            },
        };
    };

    private interruptComponentName = (
        interruptName: string,
    ): MyComponentNames | undefined => {
        if (interruptName === "APIE") {
            return "apienSctrla";
        } else if (interruptName === "DIE") {
            return "dienSctrla";
        } else if (interruptName === "PIE") {
            return "pienSctrla";
        } else if (interruptName === "RIE") {
            return "rienMctrla";
        } else if (interruptName === "WIE") {
            return "wienMctrla";
        } else {
            return undefined;
        }
    };

    private isMasterMode = (): boolean => {
        return this.dataModel.getComponentValue("mode") === "Host";
    };

    private isMasterModeInterrupt = (): boolean => {
        return (
            this.dataModel.isComponentValue("interruptDriven") &&
            this.dataModel.getComponentValue("mode") === "Host"
        );
    };

    private isMasterModePolling = (): boolean => {
        return (
            !this.dataModel.isComponentValue("interruptDriven") &&
            this.dataModel.getComponentValue("mode") === "Host"
        );
    };

    private isSlaveModeInterrupt = (): boolean => {
        return (
            this.dataModel.isComponentValue("interruptDriven") &&
            this.dataModel.getComponentValue("mode") === "Client"
        );
    };

    private shouldGenerateMasterExample = (): boolean => {
        return (
            this.dataModel.getComponentValue("mode") === "Host" &&
            this.isAssignedExport("i2c", i2c_master.Interface) === false
        );
    };

    private enableMctrla = (): string | undefined => {
        return this.boolToReg(
            this.dataModel.isComponentValue("enableI2C") &&
                this.dataModel.getComponentValue("mode") === "Host",
        );
    };

    private enableSctrla = (): string | undefined => {
        return this.boolToReg(
            this.dataModel.isComponentValue("enableI2C") &&
                this.dataModel.getComponentValue("mode") === "Client",
        );
    };

    private mbaud = (): number => {
        return this.getCalc().getMBAUD();
    };

    private actualClockSpeed = (): number => {
        return this.getCalc().getActualClockSpeed();
    };

    private getMyPinData = (): any => {
        return {};
    };

    private getMyAlerts = (): Processor.Alert[] => {
        const alerts: Processor.Alert[] = [];

        alerts.push({
            text: "Dual Mode driver is not supported in MCC.",
            type: Processor.AlertTypes.Hint,
        });

        if (
            this.dataModel.isComponentValue("fmpenCtrla") &&
            Number(this.dataModel?.getImportValue("osc_clocks")?.fosc ?? 0) < 16e6
        ) {
            alerts.push({
                text:
                    "FMP mode is selected.The Peripheral Clock (CLK_PER) has to run at 16MHz or faster.",
                type: Processor.AlertTypes.Warning,
            });
        }

        alerts.push(...(this.getModeFunctions()?.getAlerts() ?? []));

        return alerts;
    };
    private isAssignedExport = (key: MyExportNames, id: Interface): boolean => {
        const exportInterface = this.dataModel
            .getExportInterfaces()
            .getInterface(key, id.getInterfaceId());
        return exportInterface?.args != undefined;
    };

    private desiredClockSpeedValidator = (): any => {
        return {
            minimum: 1,
            maximum: Number(this.dataModel.getImportValue("osc_clocks")?.fosc) ?? 1000000,
            multipleOf: 1,
        };
    };

    private slaveAddressValidator = (): any => {
        return {
            minimum: 0,
            maximum: 127,
            multipleOf: 1,
        };
    };

    private boolToReg = (enable: boolean): string | undefined => {
        if (enable != undefined) {
            return enable ? "enabled" : "disabled";
        }
    };

    private i2c_host_basic_payload = (): i2c_host_basic.ProcessedPayload => {
        return {
            moduleName: this.getModel().getName(),
            moduleInstance: Number(
                this.getModel().getPeripheralDescription()?.instance ?? 1,
            ),
            calculatedSpeed_Hz:
                this.getModel().getComponentValue("actualClockSpeed")?.toString() ?? "0",
            i2cHostCallbackEvent: {
                header: "",
                name: "",
            },
            interruptDriven: this.getModel().isComponentValue("interruptsEnable"),
            hasInterrupts: true,
            interfaceApi: i2c_host_basic.Interface.createFirmwareApi(
                this.getModel().getComponentValue("componentName"),
            ),
        };
    };
}
