import * as i2c_master from "@microchip/i2c-host";
import * as I2CSlaveAVR8 from "./interfaces/I2CSlaveAVR8";
import { FirmwareAPIPayload, HeaderFile, buildSimpleApi } from "@microchip/scf-interface";
import { AppModel } from "@microchip/scf-automodule-interface/lib/src/types";

export const getGeneratorModel = (model: AppModel): any => {
    const interrupts = model.getHardware()?.getPeripheral().interrupts;
    const registers = model.getRegisterValues();
    const moduleName = model.getName();
    const componentName = model.getComponentValue("componentName");

    const i2cInterface =
        model.getComponentValue("mode") === "Client"
            ? i2cSlaveInterface(model)
            : i2cMasterInterface(model);

    const getFoscValue = (): string => {
        const fosc = Number(model?.getImportValue("osc_clocks")?.fosc ?? 0);
        return fosc.toString();
    };

    return {
        productName: "",
        productVersion: "",
        selectedDevice: "",
        compiler: "XC8 v2.20 and above",
        tool: "MPLABX v5.40  and above",

        moduleNameLowerCase: moduleName.toLowerCase(),
        moduleNameUpperCase: moduleName.toUpperCase(),
        moduleGroupNameLowerCase: getModuleGroupName().toLowerCase(),
        moduleGroupNameUpperCase: getModuleGroupName().toUpperCase(),
        moduleInstance: moduleName,
        componentName: componentName,
        componentNameLowerCase: componentName.toLowerCase(),
        componentNameUpperCase: componentName.toUpperCase(),
        header: i2cInterface?.headerFiles?.[0],
        api: i2cInterface.api,
        simpleApi: i2cInterface.simpleApi,
        interface: i2cInterface,
        instanceNumber: model.getPeripheralDescription()?.instance,

        CLK_PER: getFoscValue(),
        SCL_Frequency: String(model.getComponentValue("actualClockSpeed")),
        Trise: 0,

        //Registers and Settings
        initRegisters: registers,
        initInterrupts: interrupts,

        ...registers,
        ...interrupts,
    };
};

export const getHeaderFiles = (model: AppModel): HeaderFile[] => {
    return [
        {
            name: `${getFilePrefix(model)}.h`,
            path:
                model.getComponentValue("mode") === "Host" ? "i2c_host/" : "i2c_client/",
        },
    ];
};

const getFilePrefix = (model: AppModel): string => {
    return (model.getHardware()?.getName() ?? "I2C").toLowerCase();
};

export const getI2CMasterTypes = (model: AppModel): i2c_master.Typedefs => {
    const enumPrefix = getHardwareName(model).toUpperCase().replace("TWI", "I2C");
    const typePrefix = getHardwareName(model)
        .toLowerCase()
        .toLowerCase()
        .replace("twi", "i2c");
    return {
        address: typePrefix + "_address_t",
        operations: typePrefix + "_operations_t",
        error: typePrefix + "_error_t",
        callback: typePrefix + "_callback_t",
        restartWrite:
            model.getComponentValue("componentName") + "_SetRestartWriteCallback",
        BUSY: enumPrefix + "_BUSY",
        FAIL: enumPrefix + "_FAIL",
        NOERR: enumPrefix + "_NOERR",
        continue: enumPrefix + "_CONTINUE",
        restart_read: enumPrefix + "_RESTART_READ",
    };
};

export const getI2CMasterCustomApi = (model: AppModel): any => {
    const prefix = model.getComponentValue("componentName");
    const types = getI2CMasterTypes(model);
    const i2cApi = i2c_master.createApi(prefix, types);

    return {
        ...i2cApi,
        Initialize: {
            name: prefix + "_Initialize",
            returns: "uint8_t",
        },
        Open: {
            name: prefix + "_Open",
            returns: types.error,
            arguments: [
                {
                    name: "address",
                    type: types.address,
                },
            ],
        },
        Close: {
            name: prefix + "_Close",
            returns: types.error,
        },
        SetBuffer: {
            name: prefix + "_SetBuffer",
            returns: "void",
            arguments: [
                {
                    name: "*buffer",
                    type: "void",
                },
                {
                    name: "bufferSize",
                    type: "size_t",
                },
            ],
        },
        SetTimeout: {
            name: prefix + "_SetTimeout",
            returns: "void",
            arguments: [
                {
                    name: "to",
                    type: "uint8_t",
                },
            ],
        },
        SetAddressNackCallback: {
            name: prefix + "_SetAddressNackCallback",
            returns: "void",
            arguments: [
                {
                    name: "cb",
                    type: types.callback,
                },
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        SetDataNackCallback: {
            name: prefix + "_SetDataNackCallback",
            returns: "void",
            arguments: [
                {
                    name: "cb",
                    type: types.callback,
                },
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        SetDataCompleteCallback: {
            name: prefix + "_SetDataCompleteCallback",
            returns: "void",
            arguments: [
                {
                    name: "cb",
                    type: types.callback,
                },
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        SetWriteCollisionCallback: {
            name: prefix + "_SetWriteCollisionCallback",
            returns: "void",
            arguments: [
                {
                    name: "cb",
                    type: types.callback,
                },
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        SetTimeoutCallback: {
            name: prefix + "_SetTimeoutCallback",
            returns: "void",
            arguments: [
                {
                    name: "cb",
                    type: types.callback,
                },
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        CallbackReturnStop: {
            ...i2cApi.CallbackReturnStop,
            name: prefix + "_SetReturnStopCallback",
            returns: types.operations,
            arguments: [
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        CallbackReturnReset: {
            ...i2cApi.CallbackReturnReset,
            name: prefix + "_SetReturnResetCallback",
            returns: types.operations,
            arguments: [
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        CallbackRestartWrite: {
            ...i2cApi.CallbackRestartWrite,
            name: prefix + "_SetRestartWriteCallback",
            returns: types.operations,
            arguments: [
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        CallbackRestartRead: {
            ...i2cApi.CallbackRestartRead,
            name: prefix + "_SetRestartReadCallback",
            returns: types.operations,
            arguments: [
                {
                    name: "*funPtr",
                    type: "void",
                },
            ],
        },
        SetAddress: {
            name: prefix + "_SetAddress",
            returns: "void",
            arguments: [
                {
                    name: "address",
                    type: types.address,
                },
            ],
        },
        Poller: {
            name: prefix + "_Poller",
            returns: "void",
        },
        MasterWrite: i2cApi.HostWrite,
        MasterRead: i2cApi.HostRead,
        MasterOperation: i2cApi.HostOperation,
    };
};

export const i2cMasterInterface = (
    model: AppModel,
): FirmwareAPIPayload<i2c_master.API> => {
    const api = getI2CMasterCustomApi(model);
    return {
        api: api,
        headerFiles: getHeaderFiles(model),
        simpleApi: buildSimpleApi(api),
    };
};

export const i2cSlaveInterface = (
    model: AppModel,
): FirmwareAPIPayload<I2CSlaveAVR8.API> => {
    const api = I2CSlaveAVR8.createApi(
        model.getComponentValue("componentName"),
        getHardwareName(model).toUpperCase(),
    );
    return I2CSlaveAVR8.Interface.createFirmwareApi(
        model.getComponentValue("componentName"),
        getHeaderFiles(model),
        () => api,
    );
};

const getHardwareName = (model: AppModel): string => {
    return model.getHardware()?.getName() ?? "TWI";
};
const getModuleGroupName = (): string => {
    return "I2C";
};
