import { InterfaceId } from "@microchip/scf-common/lib/InterfaceId";
import * as Creator from "@microchip/scf-common/lib/Creator";
import * as Processor from "@microchip/scf-common/lib/Processor";
import * as InterfaceTypes from "@microchip/scf-interface";

/** The arguments to be provided to the implementation. */
export interface Arguments extends InterfaceTypes.Arguments {
    //Standard Interface
    customName?: string;
    interruptDriven?: boolean;
}

/** The payload to be passed back to the dependent module after the I2CSlaveAVR8 has been processed. */
export interface ProcessedPayload extends InterfaceTypes.ProcessedPayload {
    //Standard Interface
    interfaceApi: InterfaceTypes.FirmwareAPIPayload<API>;
    interruptDriven: boolean;
}

export interface API extends InterfaceTypes.CFunctions {
    Initialize: InterfaceTypes.CFunction;
    Open: InterfaceTypes.CFunction;
    Close: InterfaceTypes.CFunction;
    Read: InterfaceTypes.CFunction;
    Write: InterfaceTypes.CFunction;
    Enable: InterfaceTypes.CFunction;
    SendAck: InterfaceTypes.CFunction;
    SendNack: InterfaceTypes.CFunction;
    GoToUnaddressed: InterfaceTypes.CFunction;
    SetReadCallback: InterfaceTypes.CFunction;
    SetWriteCallback: InterfaceTypes.CFunction;
    SetAddressCallback: InterfaceTypes.CFunction;
    SetStopCallback: InterfaceTypes.CFunction;
    SetCollisionCallback: InterfaceTypes.CFunction;
    SetBusErrorCallback: InterfaceTypes.CFunction;
}

const getInterfaceId = (): InterfaceId => {
    return {
        name: "i2c-slave-avr8",
        version: "0.1.0",
    };
};

const createMock = (): ProcessedExportInterface => {
    return {
        interfaceId: getInterfaceId(),
        payload: {
            interfaceApi: Interface.createFirmwareApi("MyModule", [
                { name: "mymodule.h", path: "i2c_client/" },
            ]),
            interruptDriven: true,
        },
        args: {
            requestId: {
                customName: "MyModule",
            },
        },
    };
};

export const createApi = (prefix: string, typePrefix = prefix): API => {
    return {
        Initialize: {
            name: prefix + "_Initialize",
            returns: "uint8_t",
        },
        Open: {
            name: prefix + "_Open",
            returns: "void",
        },
        Close: {
            name: prefix + "_Close",
            returns: "void",
        },
        Read: {
            name: prefix + "_Read",
            returns: "uint8_t",
        },
        Write: {
            name: prefix + "_Write",
            returns: "void",
            arguments: [
                {
                    name: "data",
                    type: "uint8_t",
                },
            ],
        },
        Enable: {
            name: prefix + "_Enable",
            returns: "void",
        },
        SendAck: {
            name: prefix + "_SendAck",
            returns: "void",
        },
        SendNack: {
            name: prefix + "_SendNack",
            returns: "void",
        },
        GoToUnaddressed: {
            name: prefix + "_GotoUnaddressed",
            returns: "void",
        },
        SetReadCallback: {
            name: prefix + "_SetReadCallback",
            returns: "void",
            arguments: [
                {
                    name: "handler",
                    type: typePrefix + "_callback",
                },
            ],
        },
        SetWriteCallback: {
            name: prefix + "_SetWriteCallback",
            returns: "void",
            arguments: [
                {
                    name: "handler",
                    type: typePrefix + "_callback",
                },
            ],
        },
        SetAddressCallback: {
            name: prefix + "_SetAddressCallback",
            returns: "void",
            arguments: [
                {
                    name: "handler",
                    type: typePrefix + "_callback",
                },
            ],
        },
        SetStopCallback: {
            name: prefix + "_SetStopCallback",
            returns: "void",
            arguments: [
                {
                    name: "handler",
                    type: typePrefix + "_callback",
                },
            ],
        },
        SetCollisionCallback: {
            name: prefix + "_SetCollisionCallback",
            returns: "void",
            arguments: [
                {
                    name: "handler",
                    type: typePrefix + "_callback",
                },
            ],
        },
        SetBusErrorCallback: {
            name: prefix + "_SetBusErrorCallback",
            returns: "void",
            arguments: [
                {
                    name: "handler",
                    type: typePrefix + "_callback",
                },
            ],
        },
    };
};

/** The type of the processed export interface */
export type ProcessedExportInterface = Processor.ProcessedInterface<
    ProcessedPayload,
    Arguments
>;

/** Used by implementations of this interface in their module's creator. */
type CreatorExportInterface = Creator.Interface<ProcessedPayload>;

const createPrototypeImport = (): Creator.Import => {
    return {
        interfaceId: getInterfaceId(),
    };
};

const createPrototypeExport = (): Creator.Export<CreatorExportInterface[]> => {
    return {
        interfaces: [{ interfaceId: getInterfaceId() }],
    };
};

const createFirmwareApi = (
    moduleName?: string,
    headerFiles?: InterfaceTypes.HeaderFile[],
    createCustomApi?: () => API,
): InterfaceTypes.FirmwareAPIPayload<API> => {
    const prefix: string = moduleName ? moduleName.toUpperCase() : "undefined";
    const api = createCustomApi ? createCustomApi() : createApi(prefix);

    return {
        api: api,
        simpleApi: InterfaceTypes.buildSimpleApi(api),
        headerFiles: headerFiles
            ? headerFiles
            : [{ name: prefix.toLowerCase() + ".h", path: "include/" }],
    };
};

interface Interface extends InterfaceTypes.Interface {
    createPrototypeImport: () => Creator.Import<Arguments>;
    createMock: () => ProcessedExportInterface;
    createFirmwareApi: (
        moduleName?: string,
        headerFiles?: InterfaceTypes.HeaderFile[],
        createCustomApi?: () => API,
    ) => InterfaceTypes.FirmwareAPIPayload<API>;
}

export const Interface: Interface = {
    createPrototypeExport: createPrototypeExport,
    createPrototypeImport: createPrototypeImport,
    getInterfaceId: getInterfaceId,
    createMock: createMock,
    createFirmwareApi: createFirmwareApi,
};
